<?php // Please do not edit this file. If you do, save as UTF-8 without BOM!

/* Copyright 2019, 2020 Martin Janecke <mouse@prlbr.de>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

// declare (strict_types = 1); // only during development

namespace prlbr\mouse;


// Sets named x-module parameters for backwards compatibility with Aneamal 27.
function aneamal_v28_parameters (
	array &$_
):	void
{
	if (!isset ($_['base'])):
		$_['base']	= $_[3];
		$_['here']	= $_[2];
		$_['home']	= $_[0][4];
		$_['lang']	= '';
		$_['root']	= $_[0][3];
		$_['clue']	= $_[1];
		$_['meta']	= NULL;
		$_['files']	= [$_[0][1]];
		$_['hrefs']	= [$_[0][2]];
		$_['links']	= [$_[0][0]];
	endif;
}


// Returns the local file path of the crawler, including the id as parameter.
function crawler_file (
	string $id
):	string
{
	return esc (dirname (__DIR__) . '/crawl.php ' . $id);
}


// Returns an Aneamal meta javascript declaraion to trigger the crawler.
function crawler_aneamal (
	string $module_url,
	string $id
):	string
{
	return esc ('@javascript: ->' . $module_url . '/crawl.php?' . $id);
}


// Returns a HTML link to the crawler to be used by web-based cronjobs.
function crawler_link (
	string $module_url,
	string $id
):	string
{
	$url = $module_url . '/crawl.php?' . $id;
	if (
		isset ($_SERVER['REQUEST_SCHEME'])
		and isset ($_SERVER['HTTP_HOST'])
		and $_SERVER['REQUEST_SCHEME']
		and $_SERVER['HTTP_HOST']
	):
		$url = $_SERVER['REQUEST_SCHEME'] . '://' . $_SERVER['HTTP_HOST'] . $url;
	endif;
	$url = esc ($url);
	return "<a href='$url'>$url</a>";
}


// Converts characters with special function in HTML to HTML entities.
function esc (
	string $string
):	string
{
	return htmlspecialchars ($string, ENT_SUBSTITUTE | ENT_QUOTES, 'UTF-8');
}


// Returns the HTML code for the search form.
function form (
	string $action,
	string $button,
	string $name = 'x-mouse-1',
	string $term = ''
):	string
{
	$action = esc ($action);
	$button = esc ($button);
	$name = esc ($name);
	$term = esc ($term);

	$form = ['<form'];
	$action !== '' and $form[] = " action='$action'";
	$form[] = " method='post' accept-charset='UTF-8' role='search' class='x-mouse-form'>";
	$form[] = "<input name='$name' type='text' role='searchbox' maxlength='2048' value='$term' required>";
	$form[] = "<button type='submit'";
	$button !== '' and $form[] = " aria-label='$button'";
	$form[] = "><span>&#x1F50D;</span></button>";
	$form[] = "</form>";

	return implode ($form);
}


// Normalizes a set of &$urls that define the realm of the search engine and
// returns a short hash to identify that realm.
function realm (
	array &$urls
):	string
{
	// normalize and sort the homepages to get a hash independet of the provided
	// order and things like lettercase of case-insensitve URL components
	require_once __DIR__ . '/word.php';
	foreach ($urls as &$url):
		$url = word::normalize_url ($url);
	endforeach;
	sort ($urls, SORT_STRING | SORT_FLAG_CASE);

	// return an 8-character long base-32 lowercase hash that avoids the chars
	// 0, 1 and l since the latter two can be mistaken for eachother and 0 can
	// be mistaken for a big o by someone who does not know it is lowercase
	$md5 = md5 ("MOUSE\t" . implode ("\t", $urls));
	return str_replace (['0', '1', 'l'], ['x', 'y', 'z'], base_convert (substr ($md5, 0, 5), 16, 32)
		.  str_pad (base_convert (substr ($md5, -5), 16, 32), 4, '0', STR_PAD_LEFT)
	);
}


// Returns a $text marked up with HTML and CSS.
function message (
	string $text
):	string
{
	return "<div style='background:yellow;color:#000;padding:1ex 1em'>"
		.  esc ($text)
		.  "</div>"
	;
}


// Parses a comma-separated $string like 'foo=9, bar' and returns an array like
// ['foo' => '9', 'bar' => 'bar']
function parameters (
	string $string = NULL
):	array
{
	$array = [];
	if (isset ($string)):
		foreach (array_map ('trim', explode (',', $string)) as $parameter):
			if (strpos ($parameter, '=')):
				list ($key, $value) = explode ('=', $parameter, 2);
				$array[rtrim ($key)] = ltrim ($value);
			elseif ($parameter !== ''):
				$array[$parameter] = $parameter;
			endif;
		endforeach;
	endif;
	return $array;
}


// Returns HTML style element with contents of style.css in a given file path.
function style (
):	string
{
	static $included = false;
	if ($included):
		return '';
	endif;
	$included = true;
	$replace = [
		"\n"     => ' ',
		"\t"     => '',
		'{ '     => '{',
		'; }'    => '}',
		' }'     => '}',
	];
	$styles = file_get_contents (dirname (__DIR__) . '/box.css');
	return '<style>' . str_replace (array_keys ($replace), array_values ($replace), $styles) . '</style>';
}


// Returns HTML script element that triggers the crawler.
function trigger_crawler (
	string $module_url,
	string $id
):	string
{
	return "<script src='" . esc ($module_url . "/crawl.php?" . $id) . "' defer></script>";
}

// The End