<?php // Please do not edit this file. If you do, save as UTF-8 without BOM!

/* Copyright 2019, 2020 Martin Janecke <mouse@prlbr.de>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

// declare (strict_types = 1); // only during development

namespace prlbr\mouse;

require_once __DIR__ . '/phpuri/phpuri.php';


/* When doing a search, it is desired that a search term matches a word in a
 * document despite some differences, e.g. in how search term and word in the
 * document are digitally represented. We can eradicate these difference by
 * normalization, which is done by this class. This includes converting text to
 * a lowercase subset of unicode encoded as UTF-8 and removing decimal marks,
 * digit grouping symbols and leading zeros from numbers. After normalization
 * this class provides an index of words in a text:
 * $index = prlbr\mouse\word::occurrences ($text, $charset);
 * Words should be understood as actual words from a language, numbers, emojis
 * etc. Furthermore URLs can be normalized with the class, also using Phpuri by
 * P. Guardiario to turn relative URLs into absolute ones.
 */
class word {

// replacements done in a text before splitting it into words
const replace = [
	// A few white space characters are converted to a simple space character.
	// This doesn't change how words are interpreted, but we are going to use
	// the space characters to split a text into chunks for a more efficient
	// computation.
	"\r"			=> "\x20",
	"\n"			=> "\x20",
	"\t"			=> "\x20",

	// Some characters that join words like the hyphen are removed so that for 
	// example "Leinöl" and "Lein-Öl" are treated as the same term.
	"_"				=> '',  // U+005F LOW LINE (spacing underscore)
	"-"				=> '',  // U+002D HYPHEN-MINUS
	"\xC2\xAD"		=> '',  // U+00AD SOFT HYPHEN
	"\xE2\x80\x90"	=> '',	// U+2010 HYPHEN
	"\xE2\x81\xA0"	=> '',	// U+2060 WORD JOINER

	// Some characters used as decimal marks or digit grouping symbols are
	// turned into commas. The commas are treated as word separators except for
	// in numbers, where we treat them as joiners. These commas will be removed
	// at a later point and hence "1,234" matches "1234".
	"'"				=> ',',  // U+0027 APOSTROPHE
	"."				=> ',',  // U+002E FULL STOP
	"\xC2\xA0"		=> ',',  // U+00A0 NO-BREAK SPACE
	"\xE2\x80\x99"	=> ',',  // U+2019 RIGHT SINGLE QUOTATION MARK
	"\xE2\x80\xAF"	=> ',',  // U+202F NARROW NO-BREAK SPACE

	// As a concession to the Swiss, "ß" is treated like "ss".
	"ß"				=> 'ss', // U+00DF LATIN SMALL LETTER SHARP S
];


// These unreserved charaters should not be percent-encoded in URLs.
// See: https://tools.ietf.org/html/rfc3986#section-2.3
private const unreserved_characters = [
	'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'a', 'b', 'c', 'd', 'e',
	'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't',
	'u', 'v', 'w', 'x', 'y', 'z', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I',
	'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X',
	'Y', 'Z', '-', '.', '_', '~'
];


/* Methods
 */

/* Reduces a string to bytes that are in the US-ASCII range, a subset of UTF-8.
 */
private static function ascii (
	string $str
):	string
{
	$out = '';
	for ($i = 0, $len = strlen ($str); $i < $len; $i++):
		ord ($str[$i]) < 128 and $out .= $str[$i];
	endfor;
	return $out;
}


/* Turns a string into a normalized form optimized for search: UTF-8-encoded;
 * compatibility decomposed, canonical composed; lowercased; with the
 * replacements defined in self::replace applied and white space collapsed.
 */
private static function normalize (
	string $str,
	string $charset
):	string
{
	$str = mb_strtolower (\Normalizer::normalize (self::utf8 ($str, $charset), \Normalizer::FORM_KC), 'UTF-8');
	$str = trim (str_replace (array_keys (self::replace), array_values (self::replace), $str));
	do {$str = str_replace ("\x20\x20", "\x20", $str, $loop);} while ($loop);
	return $str;
}


/* Applies certain changes to an UTF-8 encoded $url:
 *
 * 1. Remove the URL fragment.
 * 2. Turn a relative URL into an absolute one using the $base URL.
 * 3. Lowercase the URL scheme and host.
 * 4. Turn an IDN host into an ACE string.
 * 5. Decode percent-encoded unreserved characters.
 * 6. Uppercase existing percent encodings.
 * 7. Percent-encode all characters that are not printable ASCII characters.
 *
 * The purposes of URL normalization are:
 *
 * <> Prevent a request for that URL from being invalid.
 * <> Prevent same URLs that are encoded differently from being handled as
 *    different URLs. Mind that URL normalization is no guarantee that two URLs
 *    which are different after normalization actually point to different
 *    ressources. Check the content of documents to see whether they differ.
 *
 * Useful reference: https://tools.ietf.org/html/rfc3987#section-5
 */
public static function normalize_url (
	string $url,
	string $base = ''
):	string
{
	// Remove URL fragment and turn the URL into an absolute one.
	$url = \phpUri::parse ($base)->join (self::strip_fragment ($url));

	// Lowercase the URL scheme.
	$scheme = parse_url ($url, PHP_URL_SCHEME) ?? '';
	$normalized_scheme = mb_strtolower ($scheme, 'UTF-8');
	if ($scheme !== '' and ($start = strpos ($url, $scheme)) !== false):
		$url = substr_replace ($url, $normalized_scheme, $start, strlen ($scheme));
	endif;

	// Lowercase the host and turn IDNs into their ASCII compatible form.
	$host = parse_url ($url, PHP_URL_HOST) ?? '';
	$normalized_host = idn_to_ascii (mb_strtolower ($host, 'UTF-8'), IDNA_DEFAULT, INTL_IDNA_VARIANT_UTS46);
	if ($host !== '' and $start = strpos ($url, $host, $start + strlen ($normalized_scheme))):
		$url = substr_replace ($url, $normalized_host, $start, strlen ($host));
	endif;

	// Decode percent-encoded unreserved characters and uppercase the other
	// remaining percent-encodings.
	$offset = 0;
	while (($pos = strpos ($url, "%", $offset)) !== false):
		$offset = $pos + 1;
		if (isset ($url[$offset]) and strspn ($hex = substr ($url, $offset, 2), '0123456789abcdefABCDEF') === 2):
			if (in_array ($byte = chr (hexdec ($hex)), self::unreserved_characters, true)):
				$url = substr_replace ($url, $byte, $pos, 3);
			else:
				$url = substr_replace ($url, strtoupper ($hex), $offset, 2);
			endif;
		endif;
	endwhile;

	// Percent-encode all bytes that are not printable ASCII characters.
	// Characters outside the ASCII range get percent-encoded from UTF-8 which
	// is the encoding expected for this method's $url parameter. Be aware that
	// some servers could expect a different encoding such as ISO-8859-1. We do
	// not know what the server actually expects and that problem cannot really
	// be solved on our side. UTF-8 as most used encoding on the WWW is our best
	// guess. (A possible but neither sure alternative to using UTF-8 would have
	// been to use the encoding which the document that contained the URL used.)
	$normalized = '';
	foreach (preg_split ('//u', $url, -1, PREG_SPLIT_NO_EMPTY) as $char):
		if (strlen ($char) > 1):
			$normalized .= '%' . implode ('%', array_map ('strtoupper', array_map ('dechex', array_map ('ord', str_split ($char)))));
		else:
			$ord = ord ($char);
			if ($ord < 33):
				$normalized .= ($ord < 16? '%0': '%') . strtoupper (dechex ($ord));
			elseif ($ord === 127):
				$normalized .= '%7F';
			else:
				$normalized .= $char;
			endif;
		endif;
	endforeach;

	return $normalized;
}


/* Returns an array whose keys are all the unique words in $str after
 * normalization and each value is an array of the places of the word in $str.
 * Word refers to a string of letters, a numerical string or a single symbol
 * such as an emoji.
 */
public static function occurrences (
	string $str,
	string $charset = 'UTF-8'
):	array
{
	$str = self::normalize ($str, $charset);

	$words = [];
	$match = [];
	$offset = 0;
	$count = 0;

	// Theoretically, we do not need this do-while-loop and could just use the
	// preg_match_all and foreach-loop. The do-while-loop feeds smaller chunks
	// of text to preg_match_all which leads to much less memory use.
	do {
		$pos = strpos ($str, "\x20", $offset);
		$chunk = $pos? substr ($str, $offset, $pos - $offset): substr ($str, $offset);
		$offset = $pos + 1;			
		if ($chunk !== '' and $chunk !== false and preg_match_all ('/\pL+|\pN+(,\pN+)*|\p{Sc}|\p{Sm}|\p{So}/u', $chunk, $match, PREG_PATTERN_ORDER)):
			foreach ($match[0] as $key => $word):
				// The next two lines only have an effect on numbers, because
				// the preg_match_all pattern above only matches commas in
				// numbers. The commas represent decimal marks and digit
				// grouping symbols to be removed. The trimming removes leading
				// zeros; but the number zero is still represented as a digit 0.
				$word = ltrim (str_replace (',', '', $word), '0');
				$word === '' and $word = '0';
				if (isset ($words[$word])):
					$words[$word][] = $key + $count;
				else:
					$words[$word] = [$key + $count];
				endif;
			endforeach;
			$count += count ($match[0]);
		endif;
	} while ($pos !== false);

	return $words;
}


/* Turns a search string into an array of search terms of type string in case of
 * single words and array in case of phrases. A phrase is marked by quotes in
 * the search string and turned into an array of words.
 */
public static function parse (
	string $string
):	array
{
	$terms = [];
	foreach (explode ('"', $string) as $n => $part):
		if ($n & 1): // $part is inside quotes
			empty ($group = self::split ($part)) or $terms[] = $group;
		else: // $part is outside quotes
			$terms = array_merge ($terms, self::split ($part));
		endif;
	endforeach;
	return $terms;
}


/* Returns an array of the words in $str after normalization. Word refers to a
 * string of letters, a numerical string or a single symbol such as an emoji.
 */
public static function split (
	string $str,
	string $charset = 'UTF-8'
):	array
{
	$str = self::normalize ($str, $charset);

	$words = [];
	$match = [];
	$offset = 0;

	// Theoretically, we do not need this do-while-loop and could just use the
	// preg_match_all and foreach-loop. The do-while-loop feeds smaller chunks
	// of text to preg_match_all which leads to much less memory use.
	do {
		$pos = strpos ($str, "\x20", $offset);
		$chunk = $pos? substr ($str, $offset, $pos - $offset): substr ($str, $offset);
		$offset = $pos + 1;			
		if ($chunk !== '' and $chunk !== false and preg_match_all ('/\pL+|\pN+(,\pN+)*|\p{Sc}|\p{Sm}|\p{So}/u', $chunk, $match, PREG_PATTERN_ORDER)):
			foreach ($match[0] as $key => $word):
				// The next two lines only have an effect on numbers, because
				// the preg_match_all pattern above only matches commas in
				// numbers. The commas represent decimal marks and digit
				// grouping symbols to be removed. The trimming removes leading
				// zeros; but the number zero is still represented as a digit 0.
				$word = ltrim (str_replace (',', '', $word), '0');
				$word === '' and $word = '0';
				$words[] = $word;
			endforeach;
		endif;
	} while ($pos !== false);

	return $words;
}


/* Returns an $url without its fragment component. This is useful when we want
 * to retrieve a document because the fragment is not part of the document's
 * address but identifies a particular spot inside a document.
 */
public static function strip_fragment (
	string $url
):	string
{
	return substr ($url, 0, strcspn ($url, '#'));
}


/* Returns an array of unique tokens with US-ASCII range bytes converted to
 * lowercase when given a $set of space-separated tokens as defined in
 * https://html.spec.whatwg.org/multipage/common-microsyntaxes.html .
 */
public static function tokens (
	string $set
):	array
{
	if ($set === ''):
		return [];
	else:
		$set = trim (str_replace (["\x09", "\x0A", "\x0C", "\x0D"], "\x20", $set));
		return array_unique (array_diff (explode ("\x20", strtolower ($set)), ['']));
	endif;
}


/* Converts a string from a given encoding into UTF-8. If the given encoding
 * is unknown to PHP or the string doesn't match its alleged encoding, then the
 * string's bytes that are in the US-ASCII range are returned.
 */
public static function utf8 (
	string $str,
	string $charset = 'UTF-8'
):	string
{
	// Check whether the encoding is known to PHP and transform the name of the
	// encoding to its preferred mime name (e.g. utf8 -> UTF-8), then check
	// whether the text matches its alleged encoding. If successful, convert to
	// UTF-8 and return the result.
	if ($charset = @mb_preferred_mime_name ($charset) and mb_check_encoding ($str, $charset)):
		$charset === 'UTF-8' or mb_convert_variables ('UTF-8', $charset, $str);
		return $str;
	endif;

	// Reduce the text to ASCII characters, if the the text didn't match its
	// alleged encoding or if the encoding is unknown.
	return self::ascii ($str);
}

} // class word

// The End