<?php // Please do not edit this file. If you do, save as UTF-8 without BOM!

/* Copyright 2019, 2020 Martin Janecke <mouse@prlbr.de>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* This module integrates a search form into an Aneamal page. It comes with and
 * triggers the crawler which builds the SQLite database to search in. The
 * database and further files are saved in the /aneamal/cache. Example use:
 *
 * [x-mouse:Find it!]->https://example.org/
 */

// declare (strict_types = 1); // only during development

namespace prlbr\mouse;

// error_reporting (E_ALL); // only during development

require_once __DIR__ . '/assets/finder.php';
require_once __DIR__ . '/assets/functions.php';
require_once __DIR__ . '/assets/i18n.php';


/* Return a closure which will be called for every instance of the search
 * module. $_ will be filled with data passed from the Aneamal core to this
 * module. $max on the other hand will not receive any value; on the contrary,
 * the declaration $max = 99 tells the Aneamal core that this module accepts
 * up to ninety-nine links to sites that shall be indexed.
 */
return function (
	array $_,
	int $max = 99
):	string
{
	// set named parameters for backwards compatibility with Aneamal < 28
	aneamal_v28_parameters ($_);

	// identificator for the current instance of the search module on a page
	// so that the correct instance responds to an HTTP post query
	static $instance = 0;
	$name = 'x-mouse-' . strval (++$instance);

	// whether the crawler shall be triggered; set to true later
	$trigger_crawler = false;

	// check all links whether they are absolute URLs; suggest an absoulte URL
	// to the homepage in case they are not
	foreach ($_['links'] as $i => $link):
		if (
			$scheme = parse_url ($link, PHP_URL_SCHEME)
			and in_array (strtolower ($scheme), ['http', 'https'], true)
			and parse_url ($link, PHP_URL_HOST)
		):
			continue;
		elseif (
			isset ($_SERVER['REQUEST_SCHEME'])
			and in_array (strtolower ($_SERVER['REQUEST_SCHEME']), ['http', 'https'], true)
			and isset ($_SERVER['HTTP_HOST'])
		):
			$home = $_SERVER['REQUEST_SCHEME'] . '://' . $_SERVER['HTTP_HOST'] . $_['home'] . '/';
			return message ('Search engine links must be absolute URLs like "' . $home . '", but "' . $link . '" is not.');
		else:
			return message ('Search engine links must be absolute URLs, but "' . $link . '" is not.');
		endif;
	endforeach;

	// start preparing the output
	$out = ["<div class='x-mouse'>"];

	// create a short hash to identify the realm of the search module that is
	// defined by the links
	$hash = realm ($_['links']);

	$cache = $_['root'] . '/aneamal/cache';
	$taskfile = $cache . '/mouse-' . $hash . '-task.txt';
	$donefile = $cache . '/mouse-' . $hash . '-done.txt';
	$dbpattern = $cache . '/mouse-' . $hash . '-*.db';

	// create a task file for the current search realm, if it does not exist yet
	if (!file_exists ($taskfile)):
		if (!is_writable ($cache)):
			return message ('Make "/aneamal/cache" writable for the search engine to work.');
		elseif (!@file_put_contents ($taskfile, implode ("\t", $_['links']) . "\n" . implode ("\n", $_['links']))):
			return message ('Setting up the search engine failed for an unknown reason.');
		endif;
		// create the donefile and date it back, so that the crawler will not
		// wait before analyzing the first page
		@touch ($donefile, time () - 30);

		$trigger_crawler = true;
	endif;

	// ARIA label for the search button
	if (isset ($_['clue']) and $_['clue'] !== ''):
		$button_label = $_['clue'];
	else:
		$button_label = i18n ($_['lang'], 'Search');
	endif;

	// the search term submitted by a user
	$term = isset ($_POST[$name])? trim ($_POST[$name]): '';

	// generate the search form
	$out[] = form ('', $button_label, $name, $term);

	// show a help page, if only a question mark is entered as search term ...
	if ($term === '?'):

		// prepare a message with the URL of the crawler
		$url = crawler_link ($_['here'], $hash);
		$script = crawler_aneamal ($_['here'], $hash);
		$out[] = "<div class='x-mouse-results'>";
		$out[] = "<dl><dt>Crawler’s URL</dt><dd><samp>$url</samp></dd>";
		$out[] = "<dt>Trigger in Aneamal</dt><dd><samp>$script</samp></dd></dl>";
		$out[] = "</div>";
		$trigger_crawler = true;

	// ... or do an actual search, if something else is entered
	elseif ($term !== ''):

		if (empty ($dbs = glob ($dbpattern))):
			$hits = [];
		else:
			// choose the second database if two exist, otherwise the first
			$db = array_pop ($dbs);
			empty ($dbs) or $db = array_pop ($dbs);
			$search = new finder ($db);
			$hits = $search->find ($term);
		endif;
		$n = count ($hits);

		// tell how many search results there are before listing the actual
		// results so that especially blind users know early what to expect
		// TODO: consider @x-mouse:results=100 as a search results limit
		$out[] = "<div class='x-mouse-results' role='region' aria-labelled-by='x-mouse-count'>";
		$out[] = "<p id='x-mouse-count' role='heading'>";
		$out[] = sprintf (i18n ($_['lang'], '%1$d search results for <i>%2$s</i>', $n), $n, esc ($term));
		$out[] = "</p>";

		// turn the results into HTML for output
		if (!empty ($hits)):
			$out[] = '<ol>';
			foreach ($hits as $url => list ($title, $score)):
				$url = esc ($url);
				$title = esc ($title);
				$score = esc ((string) $score);
				$out[] = "<li data-score='$score'><a href='$url'>$title<br><small>$url</small></a></li>";
			endforeach;
			$out[] = '</ol>';
		endif;
		$out[] = '</div>';
		$trigger_crawler = true;

	endif;

	if ($trigger_crawler):
		$out[] = trigger_crawler ($_['here'], $hash);
	endif;
	$out[] = '</div>';

	return implode ($out);
};

// Version: 1.0.1